/*
Interfacing the Arduino Duemilanove to Am29F010 devices.

Copyright (C) 2015 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

const int CE = A5, OE = A4, WE = A3, A16 = A2,
          CS1 = 2, CS2 = 3,
          DQ0 = 7, DQ1 = 6, DQ2 = 5, DQ3 = 4,
          DQ4 = 8, DQ5 = 9, DQ6 = 10, DQ7 = 11;

const int BUFSIZE = 8;
char inbuffer[BUFSIZE];
int nread = 0;

const char ERASE[][8] = {
    "W5555AA", "W2AAA55", "W555580", "W5555AA", "W2AAA55"
    };
const char PROGRAM[][8] = {
    "W5555AA", "W2AAA55", "W5555A0"
    };

int fromHex(char c)
{
    if ((c >= 48) && (c <= 57))
        return c - 48;
    if ((c >= 65) && (c <= 70))
        return c - 65 + 10;
    if ((c >= 97) && (c <= 102))
        return c - 97 + 10;
    return 0;
}

void setDataOut()
{
    pinMode(DQ0, OUTPUT);
    pinMode(DQ1, OUTPUT);
    pinMode(DQ2, OUTPUT);
    pinMode(DQ3, OUTPUT);
    pinMode(DQ4, OUTPUT);
    pinMode(DQ5, OUTPUT);
    pinMode(DQ6, OUTPUT);
    pinMode(DQ7, OUTPUT);
}

void setDataIn()
{
    pinMode(DQ0, INPUT);
    pinMode(DQ1, INPUT);
    pinMode(DQ2, INPUT);
    pinMode(DQ3, INPUT);
    pinMode(DQ4, INPUT);
    pinMode(DQ5, INPUT);
    pinMode(DQ6, INPUT);
    pinMode(DQ7, INPUT);
}

int sampleData()
{
    int i = DQ7;
    int data = 0;

    while (1)
    {
        data += digitalRead(i);
        if (i == DQ0)
            break;
        i--;
        data <<= 1;
    }

    return data;
}

int readData()
{
    int data = 0;

    setDataIn();
    delay(1);
    data = sampleData();
    setDataOut();
    return data;
}

void writeData(int data)
{
    int i = DQ0;

    while (1)
    {
        digitalWrite(i, data % 2);
        if (i == DQ7)
            break;
        i++;
        data >>= 1;
    }
}

void setAddress(int high, int low)
{
    /* The top bit is sent directly. */

    if (high & 0x100)
        digitalWrite(A16, HIGH);
    else
        digitalWrite(A16, LOW);

    /* The lower 16 bits are sent via the latches. */

    writeData(high);
    delayMicroseconds(1);
    digitalWrite(CS2, HIGH);
    delayMicroseconds(1);
    digitalWrite(CS2, LOW);
    writeData(low);
    delayMicroseconds(1);
    digitalWrite(CS1, HIGH);
    delayMicroseconds(1);
    digitalWrite(CS1, LOW);
}

int readOp(int high, int low)
{
    int data;

    digitalWrite(CE, LOW);
    setAddress(high, low);
    delay(1);
    digitalWrite(OE, LOW);
    data = readData();
    delay(1);
    digitalWrite(OE, HIGH);
    digitalWrite(CE, HIGH);

    return data;
}

void writeOp(int high, int low, int data)
{
    digitalWrite(CE, LOW);
    setAddress(high, low);
    delay(1);
    digitalWrite(WE, LOW);
    writeData(data);
    delay(1);
    digitalWrite(WE, HIGH);
    digitalWrite(CE, HIGH);
}

int readCommand(const char buffer[])
{
    int high, low, i;

    if (nread > 5)
    {
        high = fromHex(buffer[1]) << 8;
        i = 2;
    }
    else
        i = 1;

    high |= (fromHex(buffer[i]) << 4) + (fromHex(buffer[i+1]));
    low = (fromHex(buffer[i+2]) << 4) + (fromHex(buffer[i+3]));

    return readOp(high, low);
}

bool writeCommand(const char buffer[])
{
    int high, low, data, i;

    if (nread > 7)
    {
        high = fromHex(buffer[1]) << 8;
        i = 2;
    }
    else
        i = 1;

    high |= (fromHex(buffer[i]) << 4) + (fromHex(buffer[i+1]));
    low = (fromHex(buffer[i+2]) << 4) + (fromHex(buffer[i+3]));
    data = (fromHex(buffer[i+4]) << 4) + (fromHex(buffer[i+5]));

    writeOp(high, low, data);

    return true;
}

bool waitForCompletion(int high, int low, int written)
{
    int data, then = millis();

    setDataIn();
    digitalWrite(CE, LOW);
    digitalWrite(OE, LOW);

    do
    {
        delay(10);
        data = sampleData();
    } while (
        ((data & 0x80) != (written & 0x80)) &&
        ((data & 0x20) == 0) &&
        (millis() - then < 2000)
        );

    digitalWrite(OE, HIGH);
    digitalWrite(CE, HIGH);
    setDataOut();
    delay(10);
    return readOp(high, low) == written;
}

bool chipErase()
{
    for (int i = 0; i < 5; i++)
    {
        writeCommand(ERASE[i]);
    }
    writeOp(0x55, 0x55, 0x30);
    return waitForCompletion(0, 0, 0xff);
}

bool sectorErase(const char buffer[])
{
    int high;

    // 3-bit number shifted to A16...A14 with A16 discarded.
    // Thus, only sectors from 0 to 3 are supported.

    high = (fromHex(buffer[1]) << 6) & 0xff;

    for (int i = 0; i < 5; i++)
    {
        writeCommand(ERASE[i]);
    }
    writeOp(high, 0, 0x30);

    // Result of erase if 0xff written to all locations.

    return waitForCompletion(high, 0, 0xff);
}

bool program(const char buffer[])
{
    int high, low, data;

    high = (fromHex(buffer[1]) << 4) + (fromHex(buffer[2]));
    low = (fromHex(buffer[3]) << 4) + (fromHex(buffer[4]));
    data = (fromHex(buffer[5]) << 4) + (fromHex(buffer[6]));

    for (int i = 0; i < 3; i++)
    {
        writeCommand(PROGRAM[i]);
    }
    writeOp(high, low, data);

    return waitForCompletion(high, low, data);
}

void setup()
{
    Serial.begin(115200);
    pinMode(CE, OUTPUT);
    pinMode(OE, OUTPUT);
    pinMode(WE, OUTPUT);
    pinMode(A16, OUTPUT);
    pinMode(CS1, OUTPUT);
    pinMode(CS2, OUTPUT);
    setDataOut();

    // Initial state for the flash device.

    digitalWrite(CE, HIGH);
    digitalWrite(OE, HIGH);
    digitalWrite(WE, HIGH);
    digitalWrite(A16, LOW);

    // Initial state for the latches.

    digitalWrite(CS1, LOW);
    digitalWrite(CS2, LOW);

    // Interface loop.

    Serial.println("?");
}

void loop()
{
    if (nread += Serial.readBytesUntil('\n', inbuffer + nread, BUFSIZE - nread))
    {
        switch (inbuffer[0])
        {
            case 'R':
            if (nread >= 5)
            {
                Serial.println(readCommand(inbuffer), HEX);
                nread = 0;
            }
            break;

            case 'W':
            if (nread >= 7)
            {
                if (writeCommand(inbuffer))
                    Serial.println("W");
                else
                    Serial.println("C");
                nread = 0;
            }
            break;

            case 'E':
            if (chipErase())
                Serial.println("E");
            else
                Serial.println("C");
            nread = 0;
            break;

            case 'S':
            if (nread >= 2)
            {
                if (sectorErase(inbuffer))
                    Serial.println("S");
                else
                    Serial.println("C");
                nread = 0;
            }
            break;

            case 'P':
            if (nread >= 7)
            {
                if (program(inbuffer))
                    Serial.println("P");
                else
                    Serial.println("C");
                nread = 0;
            }
            break;

            default:
            Serial.println(inbuffer[0]);
            nread = 0;
            break;
        }

        Serial.flush();
    }
}

extern "C" void __cxa_pure_virtual(void) {
    while(1);
}

// tabstop=4 expandtab shiftwidth=4
